# queries to do with module/class/function relationships
from getTypeOf import getTypeOf, getTypeOfExpr
from bike.parsing.visitor import walk, walkSourceNodes
from bike import log
from bike.parsing.fastparserast import Module
import re

# first name is a child of scope
# (e.g. a method in a class, a nested function in a function, a class in a module)
def getChildByName(scope,name):
    matches = [c for c in scope.getChildNodes() if c.name == name]
    if matches != []:
        return matches[0]



def getAllClassesInHierarchy(root,classobj):
    rootClasses = _getRootClasses(classobj)
    print >>log.progress,"Getting root classes"
    relatedClasses = [] + rootClasses
    print >>log.progress,"Getting root classes - done"
    for rootClass in rootClasses:
        print >>log.progress,"Getting all derived classes for ",rootClass
        relatedClasses += getAllDerivedClasses(rootClass,root)
    return relatedClasses

def _getRootClasses(klass):
    if klass is None:  # i.e. dont have base class in our ast
        return None
    if klass.getBaseClassNames() == []:  # i.e. is a root class
        return[klass]
    else:
        rootclasses = []
        for base in klass.getBaseClassNames():
            baseclass = getTypeOf(klass,base)
            rootclass = _getRootClasses(baseclass)
            if rootclass is None:  # base class not in our ast
                rootclass = [klass]
            rootclasses+=rootclass
        return rootclasses


def getAllDerivedClasses(baseclass, root, subclasses = []):
                
    walk(root, ClassVisitor(root,baseclass,subclasses))
    return subclasses


class ClassVisitor:
    def __init__(self,root,baseclass,subclasses):
        self.baseclass = baseclass
        self.subclasses = subclasses
        self.root = root
        
    def visitSource(self,node):
        if node.containsString(self.baseclass.name):
            self.visit(node.fastparseroot)

    def visitClass(self, node):
        # create a list of possible aliases for the base class
        # (used to restrict type searches)
        baseclassname = self.baseclass.name
        baseclassNames = [baseclassname] + \
                         scanImportsForAliases(node,baseclassname)
        
        for basename in node.getBaseClassNames():
            if basename.split('.')[-1] in baseclassNames and \
                   getTypeOf(node,basename) == self.baseclass:
                self.subclasses.append(node)
                getAllDerivedClasses(node,self.root,self.subclasses)
        for child in node.getChildNodes():
            self.visit(child)


# looks for import name as foo, returns [foo,..etc..]
def scanImportsForAliases(scope,name):
    aliases = []        
    while scope:
        for lineno in scope.getImportLines():
            line = scope.module.getSourceNode().getLines()[lineno-1]
            match = re.search(name+"\s+as\s+(\w+)",line)
            if match is not None:
                aliases.append(match.group(1))
        if not isinstance(scope,Module):
            scope = scope.getParent()
        else:
            scope = None
    return aliases
